/**********************************************************************************
 * Copyright (c) 2019 Process Systems Engineering (AVT.SVT), RWTH Aachen University
 *
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License 2.0 which is available at
 * http://www.eclipse.org/legal/epl-2.0.
 *
 * SPDX-License-Identifier: EPL-2.0
 *
 **********************************************************************************/

#include "mpiUtilities.h"

#include "problem_Henry_RS_IdealGasFlash.h"
#include "problem_LP.h"
#include "problem_MILP.h"
#include "problem_NRTL_RS_Flash.h"
#include "problem_OME_RS_IdealGasFlash.h"
#include "problem_bin1.h"
#include "problem_case1_lcoe.h"
#include "problem_case2_lcoe.h"
#include "problem_case3_wnet.h"
#include "problem_ex8_1_3.h"
#include "problem_int1.h"
#include "problem_unusedVars.h"

#include "MAiNGO.h"
#include "getTime.h"

#ifdef HAVE_MAiNGO_PARSER
#include "aleModel.h"
#include "programParser.h"
#include "symbol_table.hpp"
#endif

#include <iostream>
#include <memory>
#include <string>


bool
print_testResults(std::shared_ptr<maingo::MAiNGO> theMAiNGO, maingo::RETCODE maingoStatus, const double correctSolution, const double epsilonA, const double epsilonR, double &CPUofAllProcesses)
{
    bool correctA = true;
    bool correctR = true;

#ifdef HAVE_MAiNGO_MPI
    int _rank;
    MPI_Comm_rank(MPI_COMM_WORLD, &_rank);
#endif

    MAiNGO_IF_BAB_MANAGER
        CPUofAllProcesses += theMAiNGO->get_cpu_solution_time();
        if (maingoStatus != maingo::RETCODE::GLOBALLY_OPTIMAL) {
            std::cout << "ERROR - model did not converge globally. Return code: " << maingoStatus << std::endl;
            correctA = false;
            correctR = false;
        }
        else {
            if (std::fabs(theMAiNGO->get_objective_value() - correctSolution) > epsilonA) {
                correctA = false;
				
            }
            if ((std::fabs(theMAiNGO->get_objective_value() - correctSolution)) > epsilonR * std::fabs(theMAiNGO->get_objective_value())) {
                correctR = false;
            }
            if (correctA || correctR) {
                std::cout << "OK  "
                          << " Time needed: " << std::setw(7) << std::right << theMAiNGO->get_cpu_solution_time() << "s";
                std::cout << " Time needed (wallclock): " << std::setw(7) << std::right << theMAiNGO->get_wallclock_solution_time() << "s";
                std::cout << std::endl;
            }
            else {
                if (!correctA) {
                    std::cout << "ERROR - incorrect objective value (absolute tolerance): " << theMAiNGO->get_objective_value() << " instead of " << correctSolution << std::endl;
                }
                else if (!correctR) {
                    std::cout << "ERROR - incorrect objective value (relative tolerance): " << theMAiNGO->get_objective_value() << " instead of " << correctSolution << std::endl;
                }
            }
        }
    MAiNGO_END_IF
    return (correctA || correctR);
}

bool
run_test(std::shared_ptr<maingo::MAiNGO> theMAiNGO, const std::string &name, const double correctSolution, const double epsilonA, const double epsilonR, double &CPUofAllProcesses,
         bool testAle = false, bool useMinMax = true, bool useTrig = true, bool ignoreBoundingFuncs = false, bool useRelOnly = true)
{

#ifdef HAVE_MAiNGO_MPI
    int _rank;
    MPI_Comm_rank(MPI_COMM_WORLD, &_rank);
#endif

    std::cout << std::left << std::setw(40) << name << ": ";
    const maingo::RETCODE maingoStatus                            = theMAiNGO->solve();
    MAiNGO_MPI_BARRIER
    const bool default_success = print_testResults(theMAiNGO, maingoStatus, correctSolution, epsilonA, epsilonR, CPUofAllProcesses);

#ifdef HAVE_MAiNGO_PARSER
    MAiNGO_IF_BAB_MANAGER
        theMAiNGO->write_model_to_file_in_other_language(maingo::WRITING_LANGUAGE::LANG_ALE, name + ".txt", "", useMinMax, useTrig, ignoreBoundingFuncs, useRelOnly);
    MAiNGO_END_IF
    MAiNGO_MPI_BARRIER
        std::shared_ptr<maingo::AleModel>
            myModel;
    std::ifstream input(name + ".txt");
    ale::symbol_table symbols;

    try {
        maingo::ProgramParser par(input, symbols);
        maingo::Program prog;
        par.parse(prog);
        if (par.fail()) {
            throw std::invalid_argument("Encountered an error while parsing the problem file.");
        }
        myModel = std::make_shared<maingo::AleModel>(prog, symbols);
    }
    catch (std::exception &e) {
        MAiNGO_IF_BAB_MANAGER
            std::cout << std::left << std::setw(40) << name + " (parsed)"
                      << ": ERROR - " << e.what() << std::endl;
        MAiNGO_END_IF
        MAiNGO_MPI_FINALIZE return false;
    }
    catch (...) {
        MAiNGO_IF_BAB_MANAGER
            std::cout << std::left << std::setw(40) << name + " (parsed)"
                      << ": ERROR - Encountered an unknown fatal error." << std::endl;
        MAiNGO_END_IF
        MAiNGO_MPI_FINALIZE return false;
    }

    theMAiNGO->set_model(myModel);
    std::cout << std::left << std::setw(40) << name + " (parsed)"
              << ": ";
    const maingo::RETCODE parsed_maingoStatus = theMAiNGO->solve();
    const bool parsed_success = print_testResults(theMAiNGO, parsed_maingoStatus, correctSolution, epsilonA, epsilonR, CPUofAllProcesses);
    return (default_success && parsed_success);
#else
    return default_success;
#endif

}


/**
 * @brief Main function for testing MAiNGO on given instances
 *
 * Sets options and calls the branch-and-bound solver
 */
int
main(int argc, char *argv[])
{

#ifdef HAVE_MAiNGO_MPI
    // Initialize MPI stuff
    MPI_Init(&argc, &argv);
    int _rank;
    int nProcs;
    MPI_Comm_rank(MPI_COMM_WORLD, &_rank);
    MPI_Comm_size(MPI_COMM_WORLD, &nProcs);
#endif
    int exceptionCounter = 0;
    // Define model and MAiNGO objects
    std::shared_ptr<Model_bin1> myModel_bin1;
    std::shared_ptr<Model_int1> myModel_int1;
    std::shared_ptr<Model_case1_lcoe> myModel_case1_lcoe;
    std::shared_ptr<Model_case2_lcoe> myModel_case2_lcoe;
    std::shared_ptr<Model_case3_wnet> myModel_case3_wnet;
    std::shared_ptr<Model_ex8_1_3> myModel_ex8_1_3;
    std::shared_ptr<Model_Henry_RS_IdealGasFlash> myModel_Henry_RS_IdealGasFlash;
    std::shared_ptr<Model_NRTL_RS_Flash> myModel_NRTL_RS_Flash;
    std::shared_ptr<Model_OME_RS_IdealGasFlash> myModel_OME_RS_IdealGasFlash;
    std::shared_ptr<Model_unusedVars> myModel_unusedVars;
    std::shared_ptr<Model_LP> myModel_LP;
    std::shared_ptr<Model_MILP> myModel_MILP;
    std::shared_ptr<maingo::MAiNGO> myMAiNGO;
    try {
        myModel_bin1                   = std::make_shared<Model_bin1>();                      // Define a model object implemented in problem_bin1.h
        myModel_int1                   = std::make_shared<Model_int1>();                      // Define a model object implemented in problem_int1.h
        myModel_case1_lcoe             = std::make_shared<Model_case1_lcoe>();                // Define a model object implemented in problem_case1_lcoe.h
        myModel_case2_lcoe             = std::make_shared<Model_case2_lcoe>();                // Define a model object implemented in problem_case2_lcoe.h
        myModel_case3_wnet             = std::make_shared<Model_case3_wnet>();                // Define a model object implemented in problem_case3_wnet.h
        myModel_ex8_1_3                = std::make_shared<Model_ex8_1_3>();                   // Define a model object implemented in problem_ex8_1_3.h
        myModel_Henry_RS_IdealGasFlash = std::make_shared<Model_Henry_RS_IdealGasFlash>();    // Define a model object implemented in problem_Henry_RS_IdealGasFlash.h
        myModel_NRTL_RS_Flash          = std::make_shared<Model_NRTL_RS_Flash>();             // Define a model object implemented in problem_NRTL_RS_Flash.h
        myModel_OME_RS_IdealGasFlash   = std::make_shared<Model_OME_RS_IdealGasFlash>();      // Define a model object implemented in problem_OME_RS_IdealGasFlash.h
        myModel_unusedVars             = std::make_shared<Model_unusedVars>();                // Define a model object implemented in problem_unusedVars.h
        myModel_LP                     = std::make_shared<Model_LP>();                        // Define a model object implemented in problem_LP.h
        myModel_MILP                   = std::make_shared<Model_MILP>();                      // Define a model object implemented in problem_MILP.h

        // Start with problem_bin1 and initialize MAiNGO object
        myMAiNGO = std::shared_ptr<maingo::MAiNGO>(new maingo::MAiNGO(myModel_bin1));
    }
    catch (std::exception &e) {
        exceptionCounter ++;
        MAiNGO_IF_BAB_MANAGER
            std::cerr << std::endl
                      << e.what() << std::endl;
        MAiNGO_END_IF
        MAiNGO_MPI_FINALIZE return -1;
    }
    catch (...) {
        exceptionCounter ++;
        MAiNGO_IF_BAB_MANAGER
            std::cerr << std::endl
                      << "Encountered an unknown fatal error during initialization. Terminating." << std::endl;
        MAiNGO_END_IF
        MAiNGO_MPI_FINALIZE return -1;
    }

#ifdef HAVE_MAiNGO_MPI
    // Mute cout for workers
    std::ostringstream mutestream;
    std::streambuf *coutBuf = std::cout.rdbuf();
    MAiNGO_IF_BAB_WORKER
        std::cout.rdbuf(mutestream.rdbuf());    // Note that all workers would return "ERROR - model did not converge globally."
        MAiNGO_ELSE
            std::cout
                << std::endl
                << "Running testproblems with " << nProcs << " processes (1 manager and " << nProcs - 1 << " workers)." << std::endl;
        MAiNGO_END_IF
#endif

        std::cout
            << std::endl
            << "Setting tolerances and other options." << std::endl;
        const double epsilonA = 1e-4;
        const double epsilonR = 1e-4;
        myMAiNGO->set_option("epsilonA", epsilonA);
        myMAiNGO->set_option("epsilonR", epsilonR);
        myMAiNGO->set_option("deltaIneq", 1e-6);
        myMAiNGO->set_option("deltaEq", 1e-6);
        myMAiNGO->set_option("LBP_solver", maingo::lbp::LBP_SOLVER_CPLEX);
        std::cout << "Disabling all output." << std::endl;
        myMAiNGO->set_option("loggingDestination", maingo::LOGGING_NONE);
        std::cout << std::endl;

        double CPUofAllProcesses = 0;
        const double startCPU          = maingo::get_cpu_time();
        const double startWall         = maingo::get_wall_time();
        

        // Solve the problems
        try {
            // Problem bin 1
            // The model is already set
            if (!(run_test(myMAiNGO, "Problem_bin1", 1, epsilonA, epsilonR, CPUofAllProcesses))) {
                exceptionCounter ++;
            }

            // Problem int 1
            myMAiNGO->set_model(myModel_int1);
            if (!(run_test(myMAiNGO, "Problem_int1", -0.869297426825, epsilonA, epsilonR, CPUofAllProcesses))) {
                exceptionCounter++;
            }

            // Problem case 1 lcoe
            myMAiNGO->set_model(myModel_case1_lcoe);
            if (!(run_test(myMAiNGO, "Problem_case1_lcoe", 50.2488287676, epsilonA, epsilonR, CPUofAllProcesses))) {
                exceptionCounter++;
            }

            // Problem case 2 lcoe
            myMAiNGO->set_model(myModel_case2_lcoe);
            if (!(run_test(myMAiNGO, "Problem_case2_lcoe", 48.946254947, epsilonA, epsilonR, CPUofAllProcesses))) {
                exceptionCounter++;
            }

            // Problem case 3 wnet
            myMAiNGO->set_model(myModel_case3_wnet);
            if (!(run_test(myMAiNGO, "Problem_case3_wnet", -39349.2554447061, epsilonA, epsilonR, CPUofAllProcesses))) {
                exceptionCounter++;
            }

            // Problem ex8 1 3
            myMAiNGO->set_model(myModel_ex8_1_3);
            if (!(run_test(myMAiNGO, "Problem_ex8_1_3", 3, epsilonA, epsilonR, CPUofAllProcesses))) {
                exceptionCounter++;
            }

            // Problem Henry RS Ideal Gas Flash
            myMAiNGO->set_model(myModel_Henry_RS_IdealGasFlash);
            if (!(run_test(myMAiNGO, "Problem_Henry_RS_IdealGasFlash", -0.931661985716, epsilonA, epsilonR, CPUofAllProcesses))) {
                exceptionCounter++;
            }


            // Problem NRTL RS Flash
            myMAiNGO->set_model(myModel_NRTL_RS_Flash);
            if (!(run_test(myMAiNGO, "Problem_NRTL_RS_Flash", 1064.34, epsilonA, epsilonR, CPUofAllProcesses))) {
                exceptionCounter++;
            }


            // Problem OME RS Ideal Gas Flash
            myMAiNGO->set_model(myModel_OME_RS_IdealGasFlash);
            if (!(run_test(myMAiNGO, "Problem_OME_RS_IdealGasFlash", -0.385131174192, epsilonA, epsilonR, CPUofAllProcesses))) {
                exceptionCounter++;
            }


            // Problem unused variables
            myMAiNGO->set_model(myModel_unusedVars);
            if (!(run_test(myMAiNGO, "Problem_unusedVars", 4.355812920567349, epsilonA, epsilonR, CPUofAllProcesses))) {
                exceptionCounter++;
            }


            // Problem LP
            myMAiNGO->set_model(myModel_LP);
            if (!(run_test(myMAiNGO, "Problem_LP", 153.675, epsilonA, epsilonR, CPUofAllProcesses))) {
                exceptionCounter++;
            }

            // Problem MILP
            myMAiNGO->set_model(myModel_MILP);
            if (!(run_test(myMAiNGO, "Problem_MILP", -2, epsilonA, epsilonR, CPUofAllProcesses))) {
                exceptionCounter++;
            }
        }
        catch (std::exception &e) {
            exceptionCounter ++;
            MAiNGO_IF_BAB_MANAGER
                std::cerr << std::endl
                          << e.what() << std::endl;
            MAiNGO_END_IF
            MAiNGO_MPI_FINALIZE return -1;
        }
        catch (...) {
            exceptionCounter++;
            MAiNGO_IF_BAB_MANAGER
                std::cerr << std::endl
                          << "Encountered an unknown fatal error during solution. Terminating." << std::endl;
            MAiNGO_END_IF
            MAiNGO_MPI_FINALIZE return -1;
        }
        
        if (exceptionCounter > 0) {
            throw std::exception();
            return -1;
        }

        std::cout << "Done." << std::endl
                  << std::endl;

        const double endCPU              = maingo::get_cpu_time();
        const double endWall             = maingo::get_wall_time();
        const unsigned int cpuAllMinutes = std::floor(CPUofAllProcesses / 60.);
        const unsigned int cpuMinutes    = std::floor((endCPU - startCPU) / 60.);
        const unsigned int wallMinutes   = std::floor((endWall - startWall) / 60.);
#ifdef HAVE_MAiNGO_MPI
        std::cout << "Total CPU time:  " << CPUofAllProcesses << " s = " << cpuAllMinutes << " m " << (CPUofAllProcesses)-cpuAllMinutes * (60.) << "s." << std::endl;
#else
    std::cout << "Total CPU time:  " << endCPU - startCPU << " s = " << cpuMinutes << " m " << (endCPU - startCPU) - cpuMinutes * (60.) << "s." << std::endl;
#endif
        std::cout << "Total wall time: " << endWall - startWall << " s = " << wallMinutes << " m " << (endWall - startWall) - wallMinutes * (60.) << "s." << std::endl
                  << std::endl;

#ifdef HAVE_MAiNGO_MPI
        std::cout.rdbuf(coutBuf);
#endif

        MAiNGO_MPI_FINALIZE return 0;
}
